from csv import DictWriter
from io import StringIO
from nerium.query import get_result_set


def results_to_csv(query_name, **kwargs):
    """ Generate CSV from result data
    """
    query = get_result_set(query_name, **kwargs)
    result = query.result
    stream = StringIO()
    headers = list(result[0].keys())
    dw = DictWriter(stream, fieldnames=headers)
    dw.writeheader()
    for row in result:
        dw.writerow(row)
    csv = stream.getvalue()
    return csv



# -*- coding: utf-8 -*-
"""Use SQLAlchemy engine to fetch a dataset from configured query_name
"""
import os
from sqlalchemy import create_engine


def get_data_source(query):
    """ Get data source connection metadata based on config. Prefer, in order:
        - Query file metadata
        - DATABASE_URL in environment
    """
    # from file metadata
    if "database_url" in query.metadata.keys():
        return query.metadata["database_url"]

    # Default to $DATABASE_URL (with sqlite fallback) if not set in frontmatter
    return os.getenv("DATABASE_URL", "sqlite:///")


def connection(query):
    db_url = get_data_source(query)
    db = create_engine(db_url)
    conn = db.connect()
    return conn


def result(query, **kwargs):
    try:
        db = connection(query)
        sql = query.body
        cur = db.execute(sql, **kwargs)
        cols = cur.keys()
        result = cur.fetchall()
        rows = [dict(zip(cols, row)) for row in result]
    except Exception as e:
        rows = [{"error": repr(e)}]
    return rows



def convert_multidict(obj):
    """Convert querystring params, pulling values out of list
    if there's only one.
    """
    new_dict = {
        key: (obj[key][0] if len(obj[key]) == 1 else obj[key]) for key in obj.keys()
    }
    return new_dict



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields


class ResultSchema(Schema):
    name = fields.Str()
    data = fields.List(fields.Dict(), attribute="result")
    metadata = fields.Dict()
    params = fields.Dict()
    error = fields.Str()



#!/usr/bin/env python
# -*- coding: utf-8 -*
from pathlib import Path

from dotenv import load_dotenv
from flask import Flask, jsonify, make_response, request
from flask_cors import CORS
from marshmallow import INCLUDE, Schema, fields
from marshmallow.exceptions import ValidationError
from nerium import __version__, commit, csv_result, discovery, formatter, query
from nerium.utils import convert_multidict

# Provision environment as needed
# Load local .env first
load_dotenv(Path.cwd() / ".env")
# Load this one for use w/ Kubernetes secret mount
load_dotenv("/dotenv/.env")

app = Flask(__name__)
app.url_map.strict_slashes = False
CORS(app)


@app.route("/")
@app.route("/v1/")
@app.route("/v2/")
def base_route():
    """Health check route; returns OK with version and git commit detail
    """
    return jsonify({"status": "ok", "version": __version__, "commit": commit})


@app.route("/v2/reports/")
def serve_report_list():
    """Discovery route; returns a list of available reports known to the service
    """
    return jsonify(discovery.list_reports())


@app.route("/v2/reports/<query_name>/")
def serve_report_description(query_name):
    """Discovery route; returns details and metadata about a report by name
    """
    report_descr = discovery.describe_report(query_name)
    if report_descr.error:
        status_code = getattr(report_descr, "status_code", 400)
        return jsonify(dict(error=report_descr.error)), status_code
    return jsonify(vars(report_descr))


class ResultRequestSchema(Schema):
    """Require query_name in valid results request, set format to "default" if
    not supplied
    """

    class Meta:
        unknown = INCLUDE

    query_name = fields.String(required=True)
    format_ = fields.String(missing="default", data_key="format")


def get_query_result(params):
    """Get query results from set of request parameters
    Return tuple with formatted results and status code
    """
    # load against schema, return with error message unless valid
    try:
        params = ResultRequestSchema().load(params)
    except ValidationError as e:
        return (e.normalized_messages(), 400)

    # Separate query_name and format from other parameters
    query_name = params.pop("query_name")
    format_ = params.pop("format_")

    # Fetch results from nerium.query
    query_results = query.get_result_set(query_name, **params)

    # Handle error from query submission
    if query_results.error:
        status_code = getattr(query_results, "status_code", 400)
        return (dict(error=query_results.error), status_code)

    # Format results before returning to view
    format_schema = formatter.get_format(format_)
    formatted = format_schema.dump(query_results)
    return (formatted, 200)


@app.route("/v1/<query_name>/")
@app.route("/v1/<query_name>/<format_>")
@app.route("/v2/results")
@app.route("/v2/results/<query_name>/")
@app.route("/v2/results/<query_name>/<format_>")
def serve_query_result(query_name="", format_=""):
    """Parse request and hand params to get_query_result
    """
    params = request.json or convert_multidict(request.args.to_dict(flat=False))
    if query_name:
        params["query_name"] = query_name
    if format_:
        params["format_"] = format_

    query_result = get_query_result(params)

    result = query_result[0]
    status_code = query_result[1]

    return jsonify(result), status_code


@app.route("/v1/<query_name>/csv")
@app.route("/v2/results/<query_name>/csv")
def serve_csv_result(query_name):
    params = convert_multidict(request.args.to_dict(flat=False))
    query_results = csv_result.results_to_csv(query_name, **params)
    resp = make_response()
    resp.headers["content_type"] = "text/csv"
    resp.data = query_results
    return resp



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields


class ResultSchema(Schema):
    columns = fields.Function(lambda obj: list(obj.result[0].keys()))
    data = fields.Function(lambda obj: [tuple(row.values()) for row in obj.result])



import os
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import sqlparse
import yaml
from jinja2.sandbox import SandboxedEnvironment

from nerium import db


def query_file(query_name):
    """Find file matching query_name and return Path object
    """
    flat_queries = list(Path(os.getenv("QUERY_PATH", "query_files")).glob("**/*"))
    query_file = None
    query_file_match = list(filter(lambda i: query_name == i.stem, flat_queries))
    if query_file_match:
        # TODO: Warn if more than one match
        query_file = query_file_match[0]
    return query_file


def init_query(query_name):
    """Initialize query namedtuple from file name
    """
    Query = namedtuple(
        "Query",
        [
            "name",
            "executed",
            "error",
            "metadata",
            "path",
            "body",
            "result",
            "status_code",
        ],
    )
    # `nametuple.defaults` option only available in >=3.7
    # .__new__.defaults__ works with older versions
    Query.__new__.__defaults__ = (False, {}, "", "", "", 200)
    query_obj = Query(
        name=query_name,
        executed=datetime.utcnow().isoformat(),
        path=query_file(query_name),
    )
    return query_obj


def extract_metadata(query_string):
    """ Find `:meta` labeled comment in query string and load yaml from it
    """
    tokens = sqlparse.parse(query_string)[0].tokens
    # Get comment with :meta label
    try:
        meta_comment = str(
            [
                token
                for token in tokens
                if isinstance(token, sqlparse.sql.Comment) and ":meta" in str(token)
            ][0]
        )
    except IndexError:
        return {}
    # Split YAML block out of :meta comment and load it
    meta_string = meta_comment.split("---")[1]
    metadata = yaml.safe_load(meta_string)
    return metadata


def process_template(body, **kwargs):
    """Render query body using jinja2 sandbox
    TODO: Prevent variable expansion
    """
    env = SandboxedEnvironment()
    template = env.from_string(body)
    return template.render(kwargs)


def parse_query_file(query_name):
    """Parse query file and add query body and metadata.
    """
    query_obj = init_query(query_name)
    try:
        with open(query_obj.path) as f:
            query_string = f.read()

        metadata = extract_metadata(query_string)
        rendered = process_template(query_string)
        body = sqlparse.format(rendered, strip_comments=True)

        query_obj = query_obj._replace(metadata=metadata, body=body)

    # Set 404 if no file matches query_name
    except (FileNotFoundError, TypeError):
        query_obj = query_obj._replace(
            error=f"No query found matching '{query_name}'", status_code=404
        )

    return query_obj


def get_result_set(query_name, **kwargs):
    """Call parse_query_file, then submit query and return new object with result
    """
    query = parse_query_file(query_name)

    # Bail if parser captures error:
    if query.error:
        return query

    result = db.result(query, **kwargs)
    query = query._replace(result=result)

    # Set query.error in case db module captures an excecption
    if isinstance(query.result[0], dict) and "error" in query.result[0].keys():
        error = query.result[0]["error"]
        query = query._replace(error=error, status_code=400)

    return query
